@@ -34,8 +34,6 @@ enum {
 	DEBUG = 0,
 };
 
-static VALUE IO_Event_Selector_EPoll = Qnil;
-
 enum {EPOLL_MAX_EVENTS = 64};
 
 // This represents an actual fiber waiting for a specific event.
@@ -152,6 +150,7 @@ void close_internal(struct IO_Event_Selector_EPoll *selector)
 		IO_Event_Interrupt_close(&selector->interrupt);
 	}
 }
+
 static
 void IO_Event_Selector_EPoll_Type_free(void *_selector)
 {
@@ -183,7 +182,7 @@ static const rb_data_type_t IO_Event_Selector_EPoll_Type = {
 		.dsize = IO_Event_Selector_EPoll_Type_size,
 	},
 	.data = NULL,
-	.flags = RUBY_TYPED_FREE_IMMEDIATELY,
+	.flags = RUBY_TYPED_FREE_IMMEDIATELY | RUBY_TYPED_WB_PROTECTED,
 };
 
 inline static
@@ -242,7 +241,7 @@ int IO_Event_Selector_EPoll_Descriptor_update(struct IO_Event_Selector_EPoll *se
 	} else {
 		// The IO has changed, we need to reset the state:
 		epoll_descriptor->registered_events = 0;
-		epoll_descriptor->io = io;
+		RB_OBJ_WRITE(selector->backend.self, &epoll_descriptor->io, io);
 	}
 	
 	if (epoll_descriptor->waiting_events == 0) {
@@ -252,7 +251,7 @@ int IO_Event_Selector_EPoll_Descriptor_update(struct IO_Event_Selector_EPoll *se
 			epoll_descriptor->registered_events = 0;
 		}
 		
-		epoll_descriptor->io = 0;
+		RB_OBJ_WRITE(selector->backend.self, &epoll_descriptor->io, 0);
 		
 		return 0;
 	}
@@ -332,13 +331,16 @@ VALUE IO_Event_Selector_EPoll_allocate(VALUE self) {
 	struct IO_Event_Selector_EPoll *selector = NULL;
 	VALUE instance = TypedData_Make_Struct(self, struct IO_Event_Selector_EPoll, &IO_Event_Selector_EPoll_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, Qnil);
+	IO_Event_Selector_initialize(&selector->backend, self, Qnil);
 	selector->descriptor = -1;
 	selector->blocked = 0;
 	
 	selector->descriptors.element_initialize = IO_Event_Selector_EPoll_Descriptor_initialize;
 	selector->descriptors.element_free = IO_Event_Selector_EPoll_Descriptor_free;
-	IO_Event_Array_allocate(&selector->descriptors, 1024, sizeof(struct IO_Event_Selector_EPoll_Descriptor));
+	int result = IO_Event_Array_allocate(&selector->descriptors, IO_EVENT_ARRAY_DEFAULT_COUNT, sizeof(struct IO_Event_Selector_EPoll_Descriptor));
+	if (result < 0) {
+		rb_sys_fail("IO_Event_Selector_EPoll_allocate:IO_Event_Array_allocate");
+	}
 	
 	return instance;
 }
@@ -362,7 +364,7 @@ VALUE IO_Event_Selector_EPoll_initialize(VALUE self, VALUE loop) {
 	struct IO_Event_Selector_EPoll *selector = NULL;
 	TypedData_Get_Struct(self, struct IO_Event_Selector_EPoll, &IO_Event_Selector_EPoll_Type, selector);
 	
-	IO_Event_Selector_initialize(&selector->backend, loop);
+	IO_Event_Selector_initialize(&selector->backend, self, loop);
 	int result = epoll_create1(EPOLL_CLOEXEC);
 	
 	if (result == -1) {
@@ -502,13 +504,22 @@ VALUE IO_Event_Selector_EPoll_process_wait(VALUE self, VALUE fiber, VALUE _pid,
 	
 	rb_update_max_fd(descriptor);
 	
+	// `pidfd_open` (above) may be edge triggered, so we need to check if the process is already exited, and if so, return immediately, otherwise we will block indefinitely.
+	VALUE status = IO_Event_Selector_process_status_wait(pid, flags);
+	if (status != Qnil) {
+		close(descriptor);
+		return status;
+	}
+	
 	struct IO_Event_Selector_EPoll_Waiting waiting = {
 		.list = {.type = &IO_Event_Selector_EPoll_process_wait_list_type},
 		.fiber = fiber,
 		.events = IO_EVENT_READABLE,
 	};
 	
-	int result = IO_Event_Selector_EPoll_Waiting_register(selector, 0, descriptor, &waiting);
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
+	int result = IO_Event_Selector_EPoll_Waiting_register(selector, _pid, descriptor, &waiting);
 	
 	if (result == -1) {
 		close(descriptor);
@@ -567,6 +578,8 @@ VALUE IO_Event_Selector_EPoll_io_wait(VALUE self, VALUE fiber, VALUE io, VALUE e
 		.events = RB_NUM2INT(events),
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	int result = IO_Event_Selector_EPoll_Waiting_register(selector, io, descriptor, &waiting);
 	
 	if (result == -1) {
@@ -665,6 +678,8 @@ VALUE IO_Event_Selector_EPoll_io_read(VALUE self, VALUE fiber, VALUE io, VALUE b
 		.offset = offset,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	return rb_ensure(io_read_loop, (VALUE)&io_read_arguments, io_read_ensure, (VALUE)&io_read_arguments);
 }
 
@@ -761,6 +776,8 @@ VALUE IO_Event_Selector_EPoll_io_write(VALUE self, VALUE fiber, VALUE io, VALUE
 		.offset = offset,
 	};
 	
+	RB_OBJ_WRITTEN(self, Qundef, fiber);
+	
 	return rb_ensure(io_write_loop, (VALUE)&io_write_arguments, io_write_ensure, (VALUE)&io_write_arguments);
 }
 
@@ -785,24 +802,21 @@ struct timespec * make_timeout(VALUE duration, struct timespec * storage) {
 		return NULL;
 	}
 	
-	if (FIXNUM_P(duration)) {
+	if (RB_INTEGER_TYPE_P(duration)) {
 		storage->tv_sec = NUM2TIMET(duration);
 		storage->tv_nsec = 0;
 		
 		return storage;
 	}
 	
-	else if (RB_FLOAT_TYPE_P(duration)) {
-		double value = RFLOAT_VALUE(duration);
-		time_t seconds = value;
-		
-		storage->tv_sec = seconds;
-		storage->tv_nsec = (value - seconds) * 1000000000L;
-		
-		return storage;
-	}
+	duration = rb_to_float(duration);
+	double value = RFLOAT_VALUE(duration);
+	time_t seconds = value;
 	
-	rb_raise(rb_eRuntimeError, "unable to convert timeout");
+	storage->tv_sec = seconds;
+	storage->tv_nsec = (value - seconds) * 1000000000L;
+	
+	return storage;
 }
 
 static
@@ -1039,8 +1053,7 @@ VALUE IO_Event_Selector_EPoll_wakeup(VALUE self) {
 }
 
 void Init_IO_Event_Selector_EPoll(VALUE IO_Event_Selector) {
-	IO_Event_Selector_EPoll = rb_define_class_under(IO_Event_Selector, "EPoll", rb_cObject);
-	rb_gc_register_mark_object(IO_Event_Selector_EPoll);
+	VALUE IO_Event_Selector_EPoll = rb_define_class_under(IO_Event_Selector, "EPoll", rb_cObject);
 	
 	rb_define_alloc_func(IO_Event_Selector_EPoll, IO_Event_Selector_EPoll_allocate);
 	rb_define_method(IO_Event_Selector_EPoll, "initialize", IO_Event_Selector_EPoll_initialize, 1);
